from .utils import open_url, set_value, click_element, login, check_table_value
from .salesforce import get_salesforce_object_records, get_salesforce_object_records_from_query_str, get_object_fields, get_object_field_names
from .env_msapi import get_datalake_data, get_proxy_api_data, get_datalake_data_from_ex_server
from .grid import show_grid, show_salesforce_object
from .pandas_compare import convert_columns, get_diff_df



from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait
import pandas as pd

import time


def find_elements_by_text(driver, text, tag_name=''):
    elements = []
    if len(tag_name) > 0:
        elements = driver.find_elements_by_xpath(
            "//{tag_name}//*[text()='{text}']|//{tag_name}//*[@value='{text}']|//{tag_name}//*[@placeholder='{text}']".format(
                tag_name=tag_name, text=text))
    else:
        elements = driver.find_elements_by_xpath(
            "//*[text()='{0}']|//*[@value='{0}']|//*[@placeholder='{0}']".format(text))
    return elements


class WaitVisibilityLambda:
    def __init__(self, text, tag_name):
        self._text = text
        self._tag_name = tag_name

    def __call__(self, driver):
        elements = find_elements_by_text(driver, self._text, self._tag_name)
        if not elements or len(elements) == 0:
            return False
        else:
            if is_spinner_display(driver):
                return False
            return True


def find_element(driver, text, anchor_type='', tag_name=''):
    print('find_element start: ' + time.asctime(time.localtime(time.time())))
    wait = WebDriverWait(driver, 20, 0.5)
    result = wait.until(WaitVisibilityLambda(text, tag_name))
    print('find_element after awit element: ' + time.asctime(time.localtime(time.time())))

    elements = find_elements_by_text(driver, text, tag_name)
    print('find_element elements: ' + time.asctime(time.localtime(time.time())))
    # print(elements),
    if not elements or len(elements) == 0:
        return None

    max_probability_element = elements[0]
    if not anchor_type:
        return max_probability_element
    for element in elements:
        location = element.location
        print('element {0} location {1}'.format(element.text, location))
        if '左' in anchor_type:
            max_probabilty_location = max_probability_element.location
            if max_probabilty_location['x'] > location['x']:
                max_probability_element = element
        if '右' in anchor_type:
            max_probabilty_location = max_probability_element.location
            if max_probabilty_location['x'] < location['x']:
                max_probability_element = element
        if '上' in anchor_type:
            max_probabilty_location = max_probability_element.location
            if max_probabilty_location['y'] < location['y']:
                max_probability_element = element
        if '下' in anchor_type:
            max_probabilty_location = max_probability_element.location
            if max_probabilty_location['y'] > location['y']:
                max_probability_element = element

    return max_probability_element


# 查找最合适的element
# 算法：在mc_select之下 最接近的
# mc_select_input[1].send_keys(Keys.RETURN)
def find_max_probability_element(elements, anchor_element, anchor_type=''):
    max_probability_element = None
    anchor_element_location = anchor_element.location
    for element in elements:
        location = element.location
        if '下' in anchor_type:  # 查找的元素位于锚点元素的下方
            if location['y'] > anchor_element_location['y']:
                if max_probability_element:
                    max_probabilty_location = max_probability_element.location
                    if ((max_probabilty_location['y'] - anchor_element_location['y']) > (
                            location['y'] - anchor_element_location['y']) \
                            or (max_probabilty_location['x'] - anchor_element_location['x']) > (
                                    location['x'] - anchor_element_location['x'])):
                        max_probability_element = element
                else:
                    max_probability_element = element
        else:  # 默认情况查找离锚点元素最进的元素
            if max_probability_element:
                max_probabilty_location = max_probability_element.location
                if ((max_probabilty_location['y'] - anchor_element_location['y']) > (
                        location['y'] - anchor_element_location['y']) \
                        or (max_probabilty_location['x'] - anchor_element_location['x']) > (
                                location['x'] - anchor_element_location['x'])):
                    max_probability_element = element
            else:
                max_probability_element = element

    return max_probability_element


def click_element(web_driver, text, anchor_type=''):
    web_driver.implicitly_wait(1)  # seconds
    element = find_element(web_driver, text, anchor_type)
    if element:
        element.click()
    else:
        pass  # 抛出异常


def login(web_driver, username, password, env='test'):
    if env == 'test':
        web_driver.get('https://test.salesforce.com/login.jsp?pw={1}&un={0}'.format(username, password))
        click_element(web_driver, 'Continue')
    else:  # 其他环境的okta登录处理
        pass


def open_url(driver, url):
    driver.implicitly_wait(2)
    driver.get(url)


def set_value(web_driver, label_text, value, anchor_type=''):
    label_element = find_element(web_driver, label_text, anchor_type)
    print('set_value label element:')
    print(label_element)
    if label_element:
        if 'select' in anchor_type:  # 下拉选项
            select_input_elements = find_elements_by_text(web_driver, 'Search...')
            print('input elements:')
            print(select_input_elements)
            select_input_element = find_max_probability_element(select_input_elements, label_element, '下')
            # select_input_element.send_keys(value)
            select_input_element.send_keys(Keys.RETURN)
            click_element(web_driver, value, '')
        else:  # 默认找input或 textarea元素 进行录入
            parent = label_element.find_element_by_xpath("..")
            input_elements = parent.find_elements_by_xpath("//input|//textarea")
            input_element = find_max_probability_element(input_elements, label_element, '下')
            input_element.send_keys(value)


def is_spinner_display(driver):
    spinner_elements = driver.find_elements_by_xpath("//lightning-spinner")
    for spinner_element in spinner_elements:
        if spinner_element.is_displayed():
            return True
    return False;


class WaitTableVisibilityLambda:
    def __init__(self, match):
        self._match = match

    def __call__(self, driver):
        tables = pd.read_html(driver.page_source, match=self._match)
        if len(tables) > 0:
            if is_spinner_display(driver):  # 有loading
                return False
            return True
        else:
            return False


# expect_result 为True 表示希望table和record的值匹配，为False表示table需要还没有对应的record行
def check_table_value(driver, record, expect_result=True, row_index=0, table_match=None):
    column_names = []
    for name, value in record.items():
        column_names.append(name)

    driver.implicitly_wait(1)  # seconds

    element = find_element(driver, column_names[0], tag_name='table')
    assert element is not None, 'can not find table element'

    if expect_result:
        wait = WebDriverWait(driver, 20, 0.5)
        result = wait.until(WaitTableVisibilityLambda(record[column_names[0]]))
        print(result)

    match_str = '|'.join(column_names) if not table_match else '.+'
    tables = pd.read_html(driver.page_source, match=match_str)
    assert len(tables) == 1, 'table match error: ' + match_str
    error_msg = 'no row data match with row_index: {}'.format(row_index)
    html_table_pd = tables[0][column_names]
    is_match = True
    if len(html_table_pd) > row_index:
        html_table_row = html_table_pd.iloc[row_index, :]
        print(html_table_row)
        if len(html_table_row) > 0:
            for name in column_names:
                if html_table_row[name] != record[name]:
                    is_match = False
                    error_msg = "Chcek {name} error: expect value is {expect_value}, current value is {current_value}" \
                        .format(name=name, expect_value=record[name], current_value=html_table_row[name])
                    break
        else:
            is_match = False
    else:
        is_match = False

    if expect_result:
        assert is_match, error_msg
    else:
        assert not is_match, 'value exist'


import json
import requests
import pandas as pd
from urllib.parse import urlparse, parse_qs
import configparser


def get_ms_config(ini_file, tgt_svc):
    config = configparser.ConfigParser()
    config.read(ini_file)
    base_url = config[tgt_svc]['base_url']
    token_params = {
        "grant_type": "CREDENTIAL",
        "app_id": config[tgt_svc]['app_id'],
        "app_secret": config[tgt_svc]['app_secret'],
        "tgt_svc": config[tgt_svc]['tgt_svc'],
        "user_id": config[tgt_svc]['user_id']
    }
    return base_url, token_params


def getToken(base_url, params):
    body = json.dumps(params)
    headers = {
        'Content-type': 'application/json'
    }
    url = base_url + 'token'
    r = requests.post(url, data=body, headers=headers)
    # print (r.json())
    access_token = r.json().get("access_token")
    print('{url} 请求响应时间：{used_time}, access_token={access_token}' \
          .format(url=url, used_time=r.elapsed.total_seconds(), access_token=access_token))
    return access_token


def getData(access_token, url):
    headers = {
        'Content-type': 'application/json',
        'X-ENP-AUTH': access_token
    }
    # print (url)
    r = requests.get(url, headers=headers)
    print('{0} 请求响应时间：{1}'.format(url, r.elapsed.total_seconds()))
    if r.status_code != requests.codes.ok:
        print('erorr:')
        print(r.text)
    data = r.json().get('data')
    result = []
    for record in data:
        result.append(record['attributes'])

    if len(data) > 0:
        next_link = r.json().get('links').get('next')
    else:
        next_link = ''
    return result, next_link


# 获取数据湖数据的funcation
def get_datalake_data(tgt_svc, api_path, load_all=False):
    base_url, token_params = get_ms_config('envision.ini', tgt_svc)
    access_token = getToken(base_url, token_params)
    url = base_url + api_path
    result = []
    result, next_link = getData(access_token, url)
    while load_all and next_link != None:
        data, next_link = getData(access_token, next_link)
        result = result + data
    return result


# 获取挂接服务数据的function
def get_proxy_api_data(tgt_svc, api_path, return_data_path):
    base_url, token_params = get_ms_config('envision.ini', tgt_svc)
    access_token = getToken(base_url, token_params)
    url = base_url + api_path
    headers = {
        'Content-type': 'application/json',
        'X-ENP-AUTH': access_token
    }
    r = requests.get(url, headers=headers)
    if r.status_code != requests.codes.ok:
        print('erorr:')
        print(r.text)
    result = r.json().get(return_data_path)
    return result


def post_to_datalake_data_from_ex_server(access_token, url, body={}):
    headers = {
        'Content-type': 'application/json',
        'X-ENP-AUTH': access_token
    }
    # print (url)
    body = json.dumps(body)
    r = requests.post(url, headers=headers, data=body)
    print('{0} 请求响应时间：{1}'.format(url, r.elapsed.total_seconds()))
    if r.status_code != requests.codes.ok:
        print('erorr:')
        print(r.text)
    data = r.json().get('data')
    result = []
    for record in data:
        result.append(record['attributes'])

    if len(data) > 0:
        next_link = r.json().get('links').get('next')
    else:
        next_link = ''
    return result, next_link


# 获取挂接服务数据的function
def get_datalake_data_from_ex_server(tgt_svc, prnName, lakeTableName, body={}, load_all=False):
    base_url, token_params = get_ms_config('envision.ini', tgt_svc)
    access_token = getToken(base_url, token_params)
    url = base_url + 'it/lightning/exsap/api/oa/tables?prnName={prnName}&lakeTableName={lakeTableName}' \
        .format(prnName=prnName, lakeTableName=lakeTableName)
    headers = {
        'Content-type': 'application/json',
        'X-ENP-AUTH': access_token
    }

    result = []
    result, next_link = post_to_datalake_data_from_ex_server(access_token, url, body)
    while load_all and next_link != None:
        print('next_link=' + next_link)
        qs = parse_qs(urlparse(next_link).query)
        page_size = qs['page[limit]'][0]
        offset = qs['page[offset]'][0]
        ex_next_link = url + '&pageSize={page_size}&offset={offset}'.format(page_size=page_size, offset=offset)
        data, next_link = post_to_datalake_data_from_ex_server(access_token, ex_next_link)
        result = result + data
    return result



import pandas as pd

def convert_columns(df, column_dict):
    new_df = df.rename(columns=column_dict)
    new_df = new_df[column_dict.values()]
    return new_df


# 对比两个df的体系数据差异
def get_diff_df(left_df, right_df, subset, id_key='Id'):
    def report_diff(x):
        return x[0] if x[0] == x[1] else '{} → {}'.format(*x)

    left_df.fillna('', inplace=True)  # 去掉excel的空：nan
    right_df.fillna('', inplace=True)
    left_df['version'] = "left"
    right_df['version'] = "right"
    old_accts_all = set(left_df[id_key])
    new_accts_all = set(right_df[id_key])

    dropped_accts = old_accts_all - new_accts_all
    added_accts = new_accts_all - old_accts_all
    # 合并两个项目列表
    all_data = pd.concat([left_df, right_df], ignore_index=True)
    # 去除重复的项目（没有变化的项目）subset里不包含version
    changes = all_data.drop_duplicates(subset=subset, keep='last')

    # 查找duplicated的id 生成一个list
    dupe_accts = changes[changes[id_key].duplicated() == True][id_key].tolist()
    # 检索出所有在id重复列表中的row
    dupes = changes[changes[id_key].isin(dupe_accts)]

    # Pull out the old and new data into separate dataframes
    change_new = dupes[(dupes["version"] == "right")]
    change_old = dupes[(dupes["version"] == "left")]

    # Drop the temp columns - we don't need them now
    change_new = change_new.drop(['version'], axis=1)
    change_old = change_old.drop(['version'], axis=1)

    # Index on the account numbers
    change_new.set_index(id_key, inplace=True)
    change_old.set_index(id_key, inplace=True)

    # Combine all the changes together
    df_all_changes = pd.concat([change_old, change_new],
                               axis='columns',
                               keys=['left', 'right'],
                               join='outer')

    df_all_changes = df_all_changes.swaplevel(axis='columns')[change_new.columns[0:]]
    df_changed = df_all_changes.groupby(level=0, axis=1).apply(lambda frame: frame.apply(report_diff, axis=1))
    df_changed = df_changed.reset_index()

    df_removed = changes[changes[id_key].isin(dropped_accts)]
    df_added = changes[changes[id_key].isin(added_accts)]

    return df_changed, df_removed, df_added

# old = pd.read_excel(old_file_name, 'Sheet1', na_values=['NA'])
# new = pd.read_excel(new_file_name, 'Sheet1', na_values=['NA']) # reload from excel
# subset = old.columns
# df_changed, df_removed, df_added = get_diff_df(old, new, subset)